/*
/* uucheck.c
/*
/* Various checking and processing of one input part
/**/


#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "uudeview.h"
#include "uuerror.h"


static int  lastvalid = 0;
static char lastname[255];
static int  lastpart  = 0;

static char *nofname = "UNKNOWN";

/*
/* Extract a filename from the subject line. We need anything to identify
/* the name of the program for sorting. If a nice filename cannot be found, 
/* the subject line itself is used
/**/

char *GetFileName (char *subject)
{
  char *ptr = subject, *iter, *result, *part;
  int count, length=0, alflag=0;

/*
/* If this file has no subject line, assume it is the next part of the
/* previous file (this is done in PreProcessPart)
/**/

  if (subject == NULL)
    return NULL;

/*
/* If the subject starts with 'Re', it is ignored
/* REPosts are not ignored!
/**/

  if ((subject[0] == 'R' || subject[0] == 'r') &&
      (subject[1] == 'E' || subject[1] == 'e') &&
      (subject[2] != 'P' && subject[2] == 'p')) {
    uuerror = IGNORED;
    return NULL;
  }

/*
/* Ignore a "Repost" prefix of the subject line. We don't want to get
/* a file named "Repost" :-)
/**/

  if (strncmp (subject, "REPOST", 6) == 0 ||
      strncmp (subject, "Repost", 6) == 0 ||
      strncmp (subject, "repost", 6) == 0) {
    subject += 6;
    while (*subject == ' ' || *subject == ':')
      subject++;
  }
  while (*subject == ' ') subject++;

  if ((part = strstr (subject, "Part")) == NULL)
    part = strstr (subject, "part");

  if (part == subject) {
    subject += 4;
    while (*subject == ' ') subject++;
  }

/*
/* Look for two alphanumeric strings separated by a '.'
/* (That's most likely a filename)
/**/

  while (*ptr && *ptr != 0x0a && *ptr != 0x0d && ptr != part) {
    iter  = ptr;
    count = length = alflag = 0;

    while (isalnum (*iter) || *iter == '_' || *iter == '-') {
      if (isalpha (*iter))
        alflag = 1;

      count++; iter++;
    }
    if (*iter == '\0') {
      ptr    = iter;
      length = 0;
      break;
    }
    if (*iter != '.' || count > 32 || alflag == 0) {
      ptr    = iter + 1;
      length = 0;
      continue;
    }
    iter++;
    length += count + 1;
    count   = 0;
    
    while (isalnum (iter[count]) || iter[count] == '_' || iter[count] == '-')
      count++;

    if (count > 4) {
      ptr    = iter;
      length = 0;
      continue;
    }
    if (iter[count] != '.') {
      length += count;
      break;
    }

    while (isalnum (iter[count]) || iter[count] == '.' || 
	   iter[count] == '_'    || iter[count] == '-')
      count++;

    if (count < 12) {
      length += count;
      break;
    }
    ptr    = iter;
    length = 0;
  }

  if (length == 0) { /* No filename found, use subject line for ident */
    ptr = subject;

    while (!isalpha (*ptr))
      ptr++;

    while ((isalnum (ptr[length]) || ptr[length] == ' ') && ptr+length!=part)
      length++;

    if (length) {
      if (ptr[length] == '\0' || ptr[length] == 0x0a || ptr[length] == 0x0d) {
        length--;

        while ((ptr[length] == ' ' || isdigit (ptr[length])) && length > 0)
          length--;
      }
      else {
        length--;

        while (ptr[length] == ' ' && length > 0)
          length--;
      }
      length++;
    }
  }

  if (length == 0) { /* Still found nothing? We need *something*! */
    ptr    = nofname;
    length = strlen (nofname);
  }

  if ((result = (char *) malloc (length + 1)) == NULL) {
    uuerror = NOMEM;
    return NULL;
  }
    
  memcpy (result, ptr, length);
  result[length] = '\0';
    
  return result;
}

/*
/* Extract the Part Number from the subject line.
/* We look first for numbers in (#/#)'s, then for numbers in [#/#]'s
/* and then for digits that are not part of a string.
/* If we cannot find anything, assume it is the next part of the
/* previous file.
/**/

int GetPartNo (char *subject)
{
  char *ptr = subject, *iter;
  int count, length=0;

  if (subject == NULL) {
    uuerror = NOINFO;
    return -1;
  }

  if ((subject[0] == 'R' || subject[0] == 'r') && /* Ignore replies, but not */
      (subject[1] == 'E' || subject[1] == 'e') && /* reposts                 */
      (subject[2] != 'P' && subject[2] != 'p')) {
    uuerror = IGNORED;
    return -2;
  }

  while ((iter = strchr (ptr, '(')) != NULL) {
    count = length = 0; iter++;

    while (*iter == ' ' || *iter == '#')
      iter++;

    if (!isdigit (*iter)) {
      ptr = iter;
      continue;
    }
    while (isdigit (iter[length = count]))
      count++;

    if (iter[count] == '\0' || iter[count+1] == '\0') {
      iter  += count;
      length = 0;
      break;
    }
    if (iter[count] == ')')
      break;
    else
      count++;

    while (iter[count] == ' ' || iter[count] == '#' ||
           iter[count] == '/' || iter[count] == '\\')  count++;

    if (strncmp (iter + count, "of", 2) == 0 ||
        strncmp (iter + count, "Of", 2) == 0 ||
        strncmp (iter + count, "OF", 2) == 0)
      break;

    while (isdigit (iter[count]))                     count++;
    while (iter[count] == ' ')                        count++;

    if (iter[count] == ')')
      break;

    length = 0;
    ptr    = iter;
  }

  ptr = subject;

  if (length == 0) {
    while ((iter = strchr (ptr, '[')) != NULL) {
      count = length = 0; iter++;
      
      while (*iter == ' ' || *iter == '#')
	iter++;

      if (!isdigit (*iter)) {
	ptr = iter;
	continue;
      }
      while (isdigit (iter[length = count]))
	count++;

      if (iter[count] == '\0' || iter[count+1] == '\0') {
	iter  += count;
        length = 0;
	break;
      }
      if (iter[count] == ']') 
	break;
      else
	count++;

      while (iter[count] == ' ' || iter[count] == '#' ||
             iter[count] == '/' || iter[count] == '\\')  count++;

      if (strncmp (iter + count, "of", 2) == 0 ||
          strncmp (iter + count, "Of", 2) == 0 ||
          strncmp (iter + count, "OF", 2) == 0)
        break;

      while (isdigit (iter[count]))                     count++;
      while (iter[count] == ' ')                        count++;

      if (iter[count] == ']')
	break;

      length = 0;
      ptr    = iter;
    }
  }

  if (length == 0) {
    if ((iter = strstr (subject, "part")) == NULL)
      iter = strstr (subject, "Part");

    if (iter != NULL) {
      while (isalpha (*iter) || *iter == ' ' || *iter == '.' || *iter == '-')
        iter++;
      
      while (isdigit (iter[length]))
        length++;
    }
  }

  if (length == 0) {
    count = strlen(subject) - 1;
    ptr   = subject;
 
    while (count > 0) {
      if (!isdigit(ptr[count]) || isalpha(ptr[count+1]) || ptr[count+1] == '.') {
	count--;
	continue;
      }
      length = 0;

      while (count >= 0 && isdigit (ptr[count])) {
	count--; length++;
      }
      if (count >= 0 && (isalpha (ptr[count]) || ptr[count] == '/' || ptr[count] == '.')) {
        length = 0;
        continue;
      }
      count++;
      iter = ptr + count;

      if (length > 4) {
	length = 0;
	continue;
      }
      
      break;
    }
  }

  if (length == 0) {
    uuerror = NOINFO;
    return -1;
  }

  return atoi (iter);
}

/*
/* Obtain and process some information about the data.
/**/

uufile *PreProcessPart (fileread *data)
{
  uufile *result;

  if ((result = (uufile *) malloc (sizeof (uufile))) == NULL) {
    uuerror = NOMEM;
    return NULL;
  }
  if ((result->partno = GetPartNo (data->subject)) == -2) {
    free (result);
    uuerror = IGNORED;
    return NULL;
  }
  else if (result->partno > 0)
    lastvalid = 0;

  if (data->filename != NULL) {
    if ((result->filename = (char *) malloc (strlen (data->filename) + 1)) == NULL) {
      free (result);
      uuerror = NOMEM;
      return NULL;
    }
    strcpy (result->filename, data->filename);
  }
  else
    result->filename = NULL;

  if ((result->subfname = GetFileName (data->subject)) == NULL) {
    if (result->filename != NULL) {
      if ((result->subfname = (char *) malloc (strlen (result->filename) + 1)) == NULL) {
        uuerror = NOMEM;
        free (result->filename);
        free (result);
        return NULL;
      }
      strcpy (result->subfname, result->filename);
      lastvalid = 0;
    }
    else if (lastvalid) {
      if ((result->subfname = (char *) malloc (strlen (lastname) + 1)) == NULL) {
        uuerror = NOMEM;
        free (result->filename);
        free (result);
        return NULL;
      }
      strcpy (result->subfname, lastname);
    }
    else {
      uuerror = NOINFO;
      free (result);
      return NULL;
    }
  }
  else if (lastvalid && strcmp (result->subfname, lastname) != 0)
    lastvalid = 0;

  if (result->partno == -1) {
    if (lastvalid)
      result->partno = ++lastpart;
    else {
      strcpy (lastname, result->subfname);
      lastpart  = result->partno = 1;
      lastvalid = 1;
    }
  }
  result->data = data;
  result->NEXT = NULL;
  uuerror      = OK;

  return result;
}

/*
/* At this point, all files are read in and stored in the
/* "GlobalFileList". Do some checking. All parts there?
/**/

uulist *CheckGlobalList (void)
{
  char misparts[99]; int miscount=0, count;
  uulist *liter = GlobalFileList;
  short flag, part, pflag;
  uufile *fiter, *tmp;

  while (liter != NULL) {
    miscount = 0;

    if (liter->state >= ST_OK);
    else if ((fiter = liter->thisfile) == NULL)
      liter->state = ST_MISPART;
    else {
      flag = pflag = 0;

      if ((part = fiter->partno) > 1) {
        tmp = fiter;
        while (tmp) {
          if (tmp->data->begin)
            break;
          tmp = tmp->NEXT;
        }
        if (!tmp) {
          for (count=1; count<part && miscount < 98; count++)
            misparts[miscount++] = count;
        }
      }
      if (miscount == 98) {
        liter = liter->NEXT;
        continue;
      }

      if (fiter->NEXT != NULL || fiter->data->begin ||
          fiter->data->end    || fiter->data->uudet)  { pflag = 1;
        printf ("Found File '%s' Part NOs %d ", 
		(liter->filename && liter->filename[0]) ? liter->filename : liter->subfname, 
		fiter->partno);
      }

      if (fiter->data->begin) {flag |= 1; if (pflag&0x3f) printf ("begin ");}
      if (fiter->data->end)   {flag |= 2; if (pflag&0x3f) printf ("end ");}
      if (fiter->data->uudet)  flag |= 4; 
      
      fiter = fiter->NEXT;

      while (fiter != NULL) {
        printf ("%d ", fiter->partno);

	if (fiter->partno != part + 1) {
          for (count = part+1; count < fiter->partno && miscount < 98; count++) 
            misparts[miscount++] = count;
        }
        if (miscount == 98) {
          liter = liter->NEXT;
          continue;
        }
        part = fiter->partno;

	if (fiter->data->begin) {flag |= 1; printf ("begin ");}
	if (fiter->data->end)   {flag |= 2; printf ("end ");}
	if (fiter->data->uudet)  flag |= 4;
	
	fiter = fiter->NEXT;
      }
      if (miscount) {
        free (liter->misparts);
        if ((liter->misparts = (char *) malloc (miscount + 1)) != NULL) {
          memcpy (liter->misparts, misparts, miscount);
          liter->misparts[miscount] = 0;
        }
        liter->state = ST_MISPART;
      }
      else {
	if ((flag & 4) == 0)
	  liter->state = ST_NODATA;
	else if ((flag & 1) == 0)
	  liter->state = ST_NOBEGIN;
	else if ((flag & 2) == 0)
	  liter->state = ST_NOEND;
	else {
	  printf ("OK");
	  liter->state = ST_OK;
	}
      }
    }
    if (pflag&0x3f) { pflag = 0x40;
      printf ("\n");
    }
    liter = liter->NEXT;
  }
  if (pflag)
    printf ("\n");

  return GlobalFileList;
}

uulist *checkagain (uulist *this)
{
  char misparts[99]; int miscount=0, count;
  short flag=0, part;
  uufile *fiter;

  free (this->misparts);
  this->misparts = NULL;

  if ((fiter = this->thisfile) == NULL) {
    this->state = ST_MISPART;
    return this;
  }

  if ((part = fiter->partno) > 1) {
    for (count=1; count<part && miscount < 98; count++)
      misparts[miscount++] = count;
  }

  printf ("%d ", fiter->partno);

  if (fiter->data->begin) {flag |= 1; printf ("begin ");}
  if (fiter->data->end)   {flag |= 2; printf ("end ");}
  if (fiter->data->uudet)  flag |= 4; 
      
  fiter = fiter->NEXT;

  while (fiter != NULL) {
    printf ("%d ", fiter->partno);

    if (fiter->partno != part + 1) {
      for (count = part+1; count < fiter->partno && miscount < 98; count++) 
	misparts[miscount++] = count;
    }
    if (miscount == 98)
      return this;

    part = fiter->partno;

    if (fiter->data->begin) {flag |= 1; printf ("begin ");}
    if (fiter->data->end)   {flag |= 2; printf ("end ");}
    if (fiter->data->uudet)  flag |= 4;
    
    fiter = fiter->NEXT;
  }
  if (miscount) {
    if ((this->misparts = (char *) malloc (miscount + 1)) != NULL) {
      memcpy (this->misparts, misparts, miscount);
      this->misparts[miscount] = 0;
    }
    this->state = ST_MISPART;
  }
  else {
    if ((flag & 4) == 0)
      this->state = ST_NODATA;
    else if ((flag & 1) == 0)
      this->state = ST_NOBEGIN;
    else if ((flag & 2) == 0)
      this->state = ST_NOEND;
    else
      this->state = ST_OK;
  }
  return this;
}


/*****************************************************************************
 + Frank Pilhofer                          frank@informatik.uni-frankfurt.de +
 +---------------------------------------------------------------------------+
 | Department of Computer Sciences * University of Frankfurt / Main, Germany |
 *****************************************************************************/
