#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include "uudeview.h"
#include "uucheck.h"
#include "uufree.h"
#include "uunconc.h"
#include "uuerror.h"
#include "uudecomp.h"
#include "uuutil.h"
#include "uutar.h"
#include "uuio.h"

/*
 * Operational Flags
 */

int incext = 0, stdinput = 0, interact = 1;
int decomp = 0, detar    = 0;

/*
 * Global Variables
 */

uulist  *GlobalFileList = NULL;
infile  *ftoread        = NULL;
extlist *extensions     = NULL;
char *stdfile;

/*
 * In DOS, we must open the file binary, O_BINARY is defined there
 */

#ifndef __MSDOS__
#ifndef O_BINARY
#define O_BINARY 	0
#endif
#endif

int addftoread (char *file)
{
  infile *new, *iter;

  if ((new = (infile *) malloc (sizeof (infile))) == NULL) {
    uuerror = NOMEM;
    return 0;
  }
  strcpy (new->name, file);
  new->NEXT = NULL;

  if (ftoread == NULL)
    ftoread = new;
  else {
    iter = ftoread;
    while (iter->NEXT != NULL)
      iter = iter->NEXT;

    iter->NEXT = new;
  }

  return 1;
}

int addexts (char *elist)
{
  extlist *new;
  int iter;

  while (*elist) {
    if (*elist != '.') {
      elist++;
      continue;
    }
    if (*++elist == '\0')
      break;

    if ((new = (extlist *) malloc (sizeof (infile))) == NULL) {
      uuerror = NOMEM;
      return 0;
    }
    iter = 0;

    while (*elist != '.' && *elist != '\0' && iter < 7)
      new->extension[iter++] = *elist++;

    new->extension[iter] = '\0';

    new->NEXT  = extensions;
    extensions = new;
  }
  return 1;
}

int work_file (char *filename)
{
  extlist *iter = extensions;
  char *ptr;

  if ((ptr = strrchr (filename, '.')) == NULL)
    return 1;

  ptr++;

  while (iter) {
    if (strcmp (ptr, iter->extension) == 0)
      return incext ? 1 : 0;

    iter = iter->NEXT;
  }

  return incext ? 0 : 1;
}

int proc_stdin (void)
{
  static char buffer[1024];
  FILE *target;
  int bytes;

  if ((stdfile = tempnam (NULL, "uu")) == NULL) {
    uuerror = FILEERR;
    return 0;
  }

  if ((target = fopen (stdfile, "wb")) == NULL) {
    uuerror = FILEERR;
    free (stdfile);
    return 0;
  }

  while (!feof (stdin)) {
    bytes = fread (buffer, 1, 1024, stdin);

    if (bytes == 0 && !feof (stdin)) {
      uuerror = FILEERR;
      fclose (target);
      unlink (stdfile);
      free   (stdfile);
      return 0;
    }
    if (fwrite (buffer, 1, bytes, target) != bytes) {
      uuerror = FILEERR;
      fclose (target);
      unlink (stdfile);
      free   (stdfile);
      return 0;
    }
  }

  fclose (target);

  return 1;
}

int work_comline (int argc, char *argv[])
{
  int number=1;

  for (number=1; number<argc; number++) {
    if (*argv[number] != '-' && *argv[number] != '+')
      addftoread (argv[number]);
    else switch (*(argv[number] + 1)) {
    case '\0':
      if (!stdinput) {
	if (proc_stdin() == 0) {
	  printf ("Error reading standard input");
	  break;
	}
	addftoread (stdfile);
	stdinput = 1;
      }
      break;
    case 'e':
      if (number+1 < argc && *argv[number+1] == '.') {
        incext = (*argv[number] == '+') ? 1 : 0;
        addexts (argv[++number]);
      }
      break;
    case 'i':
      interact = (*argv[number] == '+') ? 1 : 0;
      break;
    case 'd':
      decomp   = (*argv[number] == '+') ? 1 : 0;
      break;
    case 't':
      detar    = (*argv[number] == '+') ? 1 : 0;
      break;
    }
  }
  if (stdinput)
    interact = 0;

  return 1;
}

void load_files (void)
{
  infile *fiter = ftoread;
  fileread *loaded;
  uufile *fload;
  FILE *datei;
  char c;

  while (fiter) {
    if ((datei = fopen (fiter->name, "rb")) == NULL) {
      printf ("ERROR: could not open %s\n", fiter->name);
      fiter = fiter->NEXT;
      continue;
    }
    while (!feof (datei)) {
      if (uuerror) {
        if (faterr[uuerror])
          printf ("ERROR: %s while reading %s\n", uuemsg[uuerror], fiter->name);

        uuerror = 0;
      }
      /* 
       * Peek file. A Linux system was reported not to detect EOF
       */
      c = fgetc (datei);
      if (feof (datei))
        break;
      else
        ungetc (c, datei);

      if ((loaded = ScanPart (datei, fiter->name)) == NULL) 
        continue;
      
      if ((fload = PreProcessPart (loaded)) == NULL) {
        killfread (loaded);
        continue;
      }
      if (!stdinput)
        printf ("Loaded '%s' %s%s%s%s %d\n",
		(loaded->subject)  ? loaded->subject  : "",
		(loaded->filename) ? loaded->filename : "",
                (loaded->begin)?" begin":"", (loaded->end)?" end":"", 
                (loaded->uudet)?" data":"", fload->partno);
    
      if (InsertPartToList (fload) == 0) {
        killfile (fload);
        continue;
      }
    }
    fiter = fiter->NEXT;
    fclose (datei);
  }
  uuerror = 0;
}

char getyn (char def) 
{
  char line[20];

  if (interact && !feof (stdin)) { fflush (stdout);
    if (fgets (line, 19, stdin) == NULL)
      return def;
  }
  else {
    printf ("%c\n", def);
    return def;
  }

  return tolower(line[0]);
}

char *indent (int depth)
{
  static char spaces[80];

  memset (spaces, ' ', depth*2);
  spaces[depth*2] = '\0';

  return spaces;
}

int process_one (uulist *iter, int depth)
{
  int ftar=istarfile(iter), fcomp=iscomp(iter);
  char line[80], command[128], tmp, *ptr1, *ptr2;
  int res, escflag; char *name = NULL;

  do {
    free (name);

    if ((name = cutdir (iter->filename)) == NULL) {
      res  = 0;
      break;
    }

    printf ("%s%s %-15s is OK.   (action) (?) ",
	    indent(depth), filemode((int)iter->mode), name);
    
    if (fcomp && decomp) {
      printf ("%c\n", 'd');
      tmp = 'd';
    }
    else if (ftar && detar) {
      printf ("%c\n", 't');
      tmp = 't';
    }
    else
      tmp = getyn ('d');

    if (tmp == 'n') {
      if (iter->NEXT == NULL)
        res = 3;
      else
        res = 1;

      break;
    }
    else if (tmp == 'c') {
      printf ("\n\n-------------------------------------------------------------------------------\n\n");
      printf ("\tYou are now using UUDEVIEW, the uudecoder by Frank Pilhofer\n\n");
      printf ("\tThe program may be copied without restrictions as long as\n");
      printf ("\t\ta) the README file is included\n");
      printf ("\t\tb) the source is not modified\n\n");
      printf ("\tIf you like uudeview, you must send the author an email. If you\n");
      printf ("\tdon't have email access, please send a postcard. You'll find the\n");
      printf ("\tAddress in the README file\n\n");
      printf ("\tYou can always find the most recent version via anonymous ftp on\n");
      printf ("\t\tftp.uni-frankfurt.de:/pub/dist/frank\n\n");
      printf ("\t\t\tFrank Pilhofer (frank@informatik.uni-frankfurt.de)\n\n");
      printf ("-------------------------------------------------------------------------------\n\n");
      continue;
    }
    else if (tmp == 'q') {
      res = 3;
      break;
    }
    else if (tmp == '?' || tmp == '\0' || tmp == '\n') {
      printf ("\n\tYou can ...\n");
      printf ("\t  - (d)ecode the file\n");
      printf ("\t  - (y)es - same as (d)\n");
      printf ("\t  - go to (n)ext file\n");
      printf ("\t  - (e)xecute a command\n");
      printf ("\t  - (l)ist textfile\n");
      printf ("\t  - (r)ename file\n");
      if (fcomp)
	printf ("\t  - (u)ncompress the file\n");
      if (ftar)
	printf ("\t  - enter the (t)ar archive\n");
      if (depth>1)
	printf ("\t  - (q)uit archive processing\n");
      else
	printf ("\t  - (q)uit program\n");
      printf ("\t  - display (c)opyright\n\n");
      continue;
    }
    if (iter->state == ST_OK) {
      if (uudecode (iter) != 1) {
	printf ("\n%s ERROR while decoding %s (%s): %s\n\n",
		indent (depth), name,
		(iter->subfname) ? iter->subfname : "",
		uuemsg[uuerror]);
	uuerror = OK;
	res     = 0;
	break;
      }
    }
    if (!(iter->state & ST_DECODED)) { /* should never come true */
      printf ("\n%s Programmer's failure in uudeview.c line %d\n\n", __LINE__);
      res = 0;
      break;
    }
    if (tmp == 'd' || tmp == 'y') {
      if (!write_file (iter)) {
	printf ("\n%s  Error writing file\n\n", indent(depth));
	uuerror = OK;
	res     = 0;
      }
      else {
	printf ("%s  File successfully written\n", indent(depth));
	res = 1;
      }

      break;
    }
    if (iter->tarblock != -1)
      if (tar2bin (iter) == NULL) {
	printf ("\n%s ERROR while untarring %s (%s): %s\n\n",
		indent (depth), name,
		(iter->subfname) ? iter->subfname : "",
		uuemsg[uuerror]);
	uuerror = OK;
	continue;
      }
    
    if (tmp == 'u' && fcomp) {
      if (uncompress (iter) == NULL) {
	printf ("\n%s ERROR while decompressing %s (%s): %s\n\n",
		indent (depth), name,
		(iter->subfname) ? iter->subfname : "",
		uuemsg[uuerror]);
	uuerror = OK;
      }
      else {
	res = 2;
	break;
      }
    }
    else if (tmp == 't' && ftar) {
      if (tarscan (iter) == NULL) {
	printf ("\n%s ERROR while detarring %s (%s): %s\n\n",
		indent (depth), name,
		(iter->subfname) ? iter->subfname : "",
		uuemsg[uuerror]);
	uuerror = OK;
      }
      else {
	res = 2;
	break;
      }
    }
    else if (tmp == 'l') {
      printf ("\nContents of file ...\n\n");
      printf ("-------------------------------------------------------------------------------\n");
      strcpy (command, "more < ");
      strcat (command, iter->binfile);
      system (command);
      printf ("\n-------------------------------------------------------------------------------\n\n");
    }
    else if (tmp == 'e') {
      printf ("\nEnter command line ($) for file: "); fflush (stdout);
      if (fgets (line, 79, stdin) == NULL)
	printf ("\n\n ERROR: Could not get Command line\n\n");
      else if (strlen (line) > 1) {
	ptr1 = line; ptr2 = command;

	while (*ptr1 && *ptr1 != '\n') {
	  if (!escflag || *ptr1 != '\\') {
	    if (*ptr1 == '$') {
	      strcpy (ptr2, iter->binfile);
	      ptr2  += strlen (iter->binfile);
	      ptr1++;
	    }
	    else
	      *ptr2++ = *ptr1++;
	  }
	  else if (escflag) {
	    *ptr2++ = *ptr1++;
	    escflag = 0;
	  }
	  else
	    escflag = 1;
	}
	*ptr2 = '\0';

	printf ("-------------------------------------------------------------------------------\n");
	system (command);
	printf ("-------------------------------------------------------------------------------\n\n");
      }
    }
    else if (tmp == 'r') {
      printf ("\nEnter new filename: "); fflush (stdout);
      if (fgets (line, 79, stdin) == NULL)
	printf ("\n\n ERROR: Could not get filename\n");
      else if (strlen (line) > 1) {
	line[strlen(line)-1] = '\0';
	free (iter->filename);
	iter->filename = _CopyLine (line, -1);
      }
      printf ("\n");
    }
    else {
      printf ("\n%s ERROR: unknown action '%c'. Enter ? for list of options.\n\n", 
	      indent (depth), tmp);
    }
  }
  while (1);

  free (name);

  return res;
}
    
int process_files (uulist *iter, int depth)
{
  char *ptr, answer, *name;
  int count, res;

  while (iter != NULL) {
    res = 0;

    if ((name = cutdir (iter->filename)) == NULL) {
      iter = iter->NEXT;
      continue;
    }

    if (!(work_file (name))) {
      if (interact)
        printf ("%s%s %-15s ignored.\n", indent (depth),
                filemode((int)iter->mode), name);
      
      iter = iter->NEXT;
      free (name);
      continue;
    }
    switch (iter->state) {
      case ST_MISPART: ptr = "Missing Part(s)";break;
      case ST_NOBEGIN: ptr = "No Begin found"; break;
      case ST_NOEND:   ptr = "No End found";   break;
      case ST_NODATA:  ptr = NULL;             break;
      case ST_OK:      ptr = "State is OK";    break;
      default:         ptr = "Unknown State";  break;
    }
    if (iter->state < ST_OK) {
      if (ptr && iter->filename)
        printf ("\n%s ERROR: File %s (%s): %s\n\n", 
                indent (depth), name,
                (iter->subfname) ? iter->subfname : "", ptr);
      iter = iter->NEXT;
      free (name);
      continue;
    }
    if (iter->state >= ST_OK) {
      switch (process_one (iter, depth)) {
      case 2:
	res = process_files (iter->branch, depth+1);
	break;
      case 3:
	return 1;
	break;
      }
    }
    free (name);
    
    if (!res)
       iter = iter->NEXT;
  }
  return 0;
}

void sighandler (int signal)
{
  printf ("\n\nReceived Signal (%d), cleaning up temp files.\n\n", signal);

  killlist (GlobalFileList);
  exit (signal);
}

void usage(void)
{
  printf ("\n  UUDEVIEW 0.3 - a nice and user-friendly uudecoder - (w) 1994 Frank Pilhofer\n");
  printf ("  usage:\n");
  printf ("    uudeview [{+-}e .ext1.ext2] [{+-}i] [file ...]\n\n");
  printf ("\t+e/-e\tInclude or exclude extensions exclusively\n");
  printf ("\t-i\tDisable interactivity (do not ask, decode everything)\n");
  printf ("\t+d\tAutomatically uncompress / gunzip files\n\n");
  printf ("  Example:\n");
  printf ("    uudeview +e .jpg.gif -i newsfile\n");
  printf ("\tThis decodes all .jpg or .gif files encountered in <newsfile>\n");
  printf ("\twithout asking.\n\n");
}

int main(int argc, char *argv[])
{
  uulist   *llist;
  char     *ptr, line[20];

  if (argc < 2) {
    usage ();
    return 1;
  }
  signal (SIGINT, sighandler);

  init_conc    ();
  work_comline (argc, argv);
  load_files   ();

  if (!stdinput && GlobalFileList != NULL)
    printf ("\n");

  llist = CheckGlobalList();

  if (!stdinput && GlobalFileList != NULL)
    printf ("\n");

  llist = smergeparts ();

  process_files (llist, 1);

  if (!stdinput && GlobalFileList != NULL)
    printf ("\n");
  if (stdinput) {
    unlink (stdfile);
    free   (stdfile);
  }

  signal   (SIGINT, SIG_DFL);
  killlist (GlobalFileList);
  killext  (extensions);
  killif   (ftoread);

  return 0;
}
