
#define ST_READ		0x00	/* Read in, but not further processed */
#define ST_MISPART	0x01	/* Missing Part detected */
#define ST_NOBEGIN	0x02	/* No 'begin' found */
#define ST_NOEND	0x03	/* No 'end' found */
#define ST_NODATA       0x05    /* File does not contain valid uudata */
#define ST_OK           0x08    /* All Parts found, ready to decode */
#define ST_DECOERR	0x10	/* Error while decoding */
#define ST_DECODED	0x20	/* Successfully decoded */
#define ST_UNCOMP	0x40	/* Decoded and uncompressed */


/*
/* Structure that holds the information for a single file / part of
/* a file. If a subject line is encountered, it is copied to subject;
/* if a begin is found, the mode and name of the file is extracted.
/* flags are set if 'begin' or 'end' is detected and 'uudet' if valid
/* uuencoded data is found. If the file contains a 'From:' line with
/* a '@' in it (indicating an origin email address), it is preserved
/* in 'origin'.
/**/

typedef struct _fileread {
  char *subject;      /* Whole subject line */
  char *filename;     /* Only filled in if begin detected */
  char *origin;       /* Whole 'From:' line */
  short mode;         /* Mode of File (from 'begin') */
  unsigned begin : 1; /* begin detected */
  unsigned end   : 1; /* end detected */
  unsigned uudet : 1; /* valid uuencoded data detected */

  char *sfname;       /* Associated source file */
  long startpos;      /* ftell() position where data starts */
  long length;        /* length of data */
} fileread;

/*
/* Structure for holding one part of a file
/* Linked List, ordered by partno.
/**/

typedef struct _uufile {
  char     *filename;
  char     *subfname;
  short     partno;
  fileread *data;
  struct _uufile *NEXT;
} uufile;

typedef struct _uulist {
  short state, mode;

  unsigned begin : 1;
  unsigned end   : 1;
  unsigned uudet : 1;

  char    *filename;
  char    *subfname;

  long     tarblock;
  char    *binfile;

  uufile  *thisfile;
  char    *misparts;

  struct _uulist *branch;
  struct _uulist *NEXT;
} uulist;

extern uulist *GlobalFileList;

typedef struct _extlist {
  char extension[8];
  struct _extlist *NEXT;
} extlist;

typedef struct _infile {
  char  name[128];
  struct _infile *NEXT;
} infile;

extern infile *ftoread;
extern extlist *extensions;

