#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#ifndef __MSDOS__
#include <unistd.h>
#endif

#include "uudeview.h"
#include "uuerror.h"
#include "uuutil.h"
#include "uutar.h"

static char ungotline[255];
static int  ungotflag  = 0;
static long oldposition= 0;
static char wrbuffer[1024];

/*
 * In DOS, we must open the file binary, O_BINARY is defined there
 */

#ifndef __MSDOS__
#ifndef O_BINARY
#define O_BINARY      0
#endif
#endif

/*
/* Load a part of a file. Search for the Subject Line (ptr), the
/* "Origin" (fptr); for begin's and end's and the filename (nptr)
/**/

fileread *ScanPart (FILE *datei, char *fname)
{
  static char line[255], *ptr;
  fileread *result;
  int subflag=0;
  int dflag=0;

  if (datei == NULL || feof (datei)) {
    uuerror = EOFILE;
    return NULL;
  }

  if ((result = (fileread *) malloc (sizeof (fileread))) == NULL) {
    uuerror = NOMEM;
    return NULL;
  }
  memset (result, 0, sizeof (fileread));

  if ((result->sfname = _CopyLine (fname, -1)) == NULL) {
    uuerror = NOMEM;
    free (result);
    return NULL;
  }
  result->startpos = (ungotflag) ? oldposition : ftell (datei);

  while (!feof (datei) || ungotflag) {
    if (ungotflag) {
      strcpy (line, ungotline);
      ungotflag = 0;
    }
    else {
      oldposition = ftell (datei);
      if (fgets (line, 255, datei) == NULL)
	break;
    }

    line[255] = '\0'; /* For Safety of string functions */

    if (strncmp (line+1, "ubject:", 7) == 0) {
      if (subflag & 1) {
	strcpy (ungotline, line);
	ungotflag = 1;
	break;
      }
      else {
	result->subject = _CopyLine (line + 8, -1);
	subflag        |= 1;
      }
    }
    if (strncmp (line+1, "rom:", 4) == 0 &&
	strchr  (line+1, '@')       != NULL) {
      if (subflag & 2) {
	strcpy (ungotline, line);
	ungotflag = 1;
	break;
      }
      else {
	result->origin = _CopyLine (line + 5, -1);
	subflag       |= 2;
      }
    }

    if (strncmp (line, "begin ", 6) == 0) {
      if (result->end) {
	strcpy (ungotline, line);
	ungotflag = 1;
	break;
      }

      ptr    = line + 5;

      while (*ptr == ' ') ptr++;
      while (isdigit (*ptr)) 
	result->mode = result->mode * 8 + *ptr++ - '0';
      while (*ptr == ' ') ptr++;

      result->filename = _CopyLine (ptr, -1);
      result->begin    = 1;
    }
    if (strncmp (line, "end", 3) == 0)
      result->end = 1;
      
    if (!(result->uudet)) {
      if (valid_data (line)) {
        if (dflag==3)
	  result->uudet = 1;
        else
          dflag++;
      }
      else
        dflag = 0;
    }
  }

  result->length = ftell (datei) - result->startpos;

  return result;
}

int write_tar (FILE *source, FILE *target, uulist *this)
{
  tarblock tb;
  long size=0;
  int bytes;

  fseek (source, this->tarblock * TARBLOCK, SEEK_SET);
  fread (&tb, TARBLOCK, 1, source);

  size = getokt (tb.header.size);

  fseek (source, (this->tarblock + 1) * TARBLOCK, SEEK_SET);

  while (size > 0) {
    bytes = fread (wrbuffer, 1, (size > 1024) ? 1024 : size, source);

    if (bytes == 0 && size != 0) {
      uuerror = FILEERR;
      fclose (source);
      fclose (target);
      unlink (this->filename);
      return 0;
    }
    if (fwrite (wrbuffer, 1, bytes, target) != bytes) {
      uuerror = FILEERR;
      fclose (source);
      fclose (target);
      unlink (this->filename);
      return 0;
    }
    size -= bytes;
  }
  return 1;
}

int write_file (uulist *this)
{
  FILE *target, *source;
  int fildes, bytes;
  char *name, *ptr;

  if ((name = cutdir (this->filename)) == NULL)
    return 0;

#ifdef __MSDOS__
  if ((ptr = dosfname (name)) == NULL) {
    free (name);
    return 0;
  }
  free (name);
  name = ptr;
#endif

  if ((source = fopen (this->binfile, "rb")) == NULL) {
    uuerror = FILEERR;
    free (name);
    return 0;
  }

  if ((fildes = open (name, O_WRONLY | O_CREAT | O_BINARY, this->mode)) == -1) {
    uuerror = FILEERR;
    fclose (source);
    free   (name);
    return 0;
  }

  if ((target = fdopen (fildes, "wb")) == NULL) {
    uuerror = FILEERR;
    fclose (source);
    close (fildes);
    free  (name);
    return 0;
  }

  if (this->tarblock >= 0) {
    if (write_tar (source, target, this) != 1) {
      uuerror = FILEERR;
      fclose (source);
      fclose (target);
      unlink (this->filename);
      free   (name);
      return 0;
    }
    fclose (source);
    fclose (target);
    free   (name);

    return 1;
  }

  while (!feof (source)) {
    bytes = fread (wrbuffer, 1, 1024, source);

    if (bytes == 0 && !feof (source)) {
      uuerror = FILEERR;
      fclose (source);
      fclose (target);
      unlink (this->filename);
      free   (name);
      return 0;
    }
    if (fwrite (wrbuffer, 1, bytes, target) != bytes) {
      uuerror = FILEERR;
      fclose (source);
      fclose (target);
      unlink (this->filename);
      free   (name);
      return 0;
    }
  }

  fclose (target);
  fclose (source);
  free   (name);

  return 1;
}


/*****************************************************************************
 + Frank Pilhofer                          frank@informatik.uni-frankfurt.de +
 +---------------------------------------------------------------------------+
 | Department of Computer Sciences * University of Frankfurt / Main, Germany |
 *****************************************************************************/
