
/*
/* These functions do the uudecoding. It is converted from a freeware
/* program named "uunconc". It didn't contain the name of the author.
/* If you are the author or know him, contact me.
/**/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

#ifndef __MSDOS__
#include <unistd.h>
#endif

#include "uudeview.h"
#include "uuerror.h"
#include "uuutil.h"

#define BEGIN     (1)
#define DATA      (2)
#define END       (3)
#define DONE      (4)

static int      xlat[256];              /* character translation table */
static int      xlen[64];               /* line length table */
static int      state = BEGIN;          /* current processing state */
static char     save[3][256];           /* buffer for decoding */

void init_conc(void)
{
  register int i, j;

  /* prepare decoding translation table */
  for(i = 0; i < ' '; i++)
    xlat[i] = -1;

  for(i = ' ', j = 0; i < ' ' + 64; i++, j++)
    xlat[i] = j;

  for(i = ' ' + 64; i < 256; i++)
    xlat[i] = -1;

  /* add special cases */
  xlat['`'] = xlat[' '];
  xlat['~'] = xlat['^'];

  /* prepare line length table */
  xlen[0] = 1;
  for(i = 1, j = 5; i <= 60; i += 3, j += 4)
    xlen[i] = xlen[i+1] = xlen[i+2] = j;
}

static int _uustrlen (char *string)
{
  int result=0;

  while (*string && *string != 0x0a && *string != 0x0d) {
    result++;
    string++;
  }
  
  if (*string == 0x0d)
    result++;
  else if (*string == 0x0a)
    result++;

  return result;
}

int valid_data(char *s)
{
  register int i, j;

  if((s == NULL) || (*s < '\0')) {
    return(0);              /* bad string */
  }
  j = xlen[xlat[*s]];
  i = _uustrlen(s);

  if((i - 1) == j)
    i--;                    /* remove trailing character */
  else if (i != j)
    return(0);              /* bad length */

  while(i--) {
    if(*s < 0 || xlat[*s++] < 0) {
      return(0);            /* bad code character */
    }
  }
  return(1);                /* data is valid */
}

int decode(char *s, char *d)
{
  int i, j, c, cc, count=0;

  i = xlat[*s++];
  j = xlen[i] - 1;

  while(j > 0) {
    c  = xlat[*s++] << 2;
    cc = xlat[*s++];
    c |= (cc >> 4);

    if(i-- > 0)
      d[count++] = c;

    cc <<= 4;
    c    = xlat[*s++];
    cc  |= (c >> 2);

    if(i-- > 0)
      d[count++] = cc;

    c <<= 6;
    c |= xlat[*s++];

    if(i-- > 0)
      d[count++] = c;

    j -= 4;
  }
  return count;
}

int uunconc (FILE *datain, FILE *dataout, int *state, long maxpos)
{
  char line[255], oline[400];
  int count, tc=0, tf=0;

  while (!feof (datain) && *state != DONE && ftell (datain) < maxpos) {
    if (fgets (line, 255, datain) == NULL)
      break;

    line[255] = '\0'; /* For Safety of string functions */
    count     =  0;

    if (*state == BEGIN) {
      if (strncmp (line, "begin ", 6) == 0)
	*state = DATA;
      continue;
    }
    else if (*state == END) {
      if (strncmp (line, "end", 3) == 0) {
	if (tc) {
	  for (tf=0; tf<tc; tf++) 
	    count += decode (save[tf], oline + count);
	}
	*state = DONE;
      }
    }
    if (*state == DATA || *state == END) {
      if (valid_data (line)) {
	if (tf)
	  count  = decode (line, oline);
	else if (tc == 3) {
	  count  = decode (save[0], oline);
	  count += decode (save[1], oline + count);
	  count += decode (save[2], oline + count);
	  count += decode (line,    oline + count);
	  tf     = 1;
	  tc     = 0;
	}
	else {
	  strcpy (save[tc++], line);
	}
	*state = (line[0] == 'M') ? DATA : END;
      }
      else {
	tf = 0;
	tc = 0;
      }
    }
    else if (*state != DONE) {
      uuerror = DECOERR;
      return -1;
    }
    if (count)
      if (fwrite (oline, 1, count, dataout) != count) {
	uuerror = FILEERR;
	return -1;
      }
  }
  if (*state == DONE)
    return 1;

  return 0;
}

int uudecode (uulist *data)
{
  FILE *datain, *dataout;
  int state = BEGIN, res;
  uufile *iter;


  if (data == NULL || data->thisfile == NULL) {
    uuerror = ILLVAL;
    return 0;
  }
  if (data->state & (ST_DECOERR | ST_DECODED)) {
    uuerror = OK;
    return 0;
  }

  if ((data->binfile = tempnam (NULL, "uu")) == NULL) {
    uuerror = NOMEM;
    return 0;
  }
  if ((dataout = fopen (data->binfile, "wb")) == NULL) {
    free (data->binfile);
    data->binfile = NULL;
    uuerror = FILEERR;
    return 0;
  }

  iter = data->thisfile;
  while (iter) {
    if (iter->data->sfname == NULL) {
      iter = iter->NEXT;
      continue;
    }
    if ((datain = fopen (iter->data->sfname, "rb")) == NULL) {
      iter = iter->NEXT;
      continue;
    }
    fseek         (datain, iter->data->startpos, SEEK_SET);
    res = uunconc (datain, dataout, &state, iter->data->startpos+iter->data->length);
    fclose        (datain);

    if (state == DONE || res == -1)
      break;
    
    iter = iter->NEXT;
  }

  fclose (dataout);

  if (state != DONE) {
    free (data->binfile);
    data->binfile = NULL;
    data->state   = ST_DECOERR;
    uuerror       = DECOERR;
    
    return 0;
  }
  data->state = ST_DECODED;

  return 1;
}
