#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#ifndef __MSDOS__
#include <unistd.h>
#endif

#include "uudeview.h"
#include "uuerror.h"
#include "uutar.h"
#include "uuutil.h"
#include "uuio.h"

uulist *tar2bin (uulist *this)
{
  FILE *datin, *datout;
  char *tname;

  if (this == NULL)
    return NULL;
  if (this->tarblock == -1)
    return NULL;

  if ((tname = tempnam (NULL, "uu")) == NULL) {
    uuerror = FILEERR;
    return NULL;
  }

  if ((datin = fopen (this->binfile, "rb")) == NULL) {
    uuerror = FILEERR;
    free (tname);
    return NULL;
  }
  if ((datout = fopen (tname, "wb")) == NULL) {
    uuerror = FILEERR;
    fclose (datin);
    free   (tname);
    return NULL;
  }
  if (write_tar (datin, datout, this) != 1) {
    uuerror = FILEERR;
    fclose (datout);
    fclose (datin);
    unlink (tname);
    free   (tname);
    return NULL;
  }
  fclose (datout);
  fclose (datin);
  free   (this->binfile);

  this->tarblock = -1;
  this->binfile  = tname;

  return this;
}

uulist *tarscan (uulist *this)
{
  uulist *new = NULL, *result = NULL, *iter = NULL;
  long fsize, blockno = 0;
  tarblock thisblock;
  FILE *tarfile;
  char *ptr;

  if (this == NULL)
    return NULL;
 
  if (this->branch != NULL)
    return this->branch;

  if ((ptr = strrchr (this->filename, '.')) == NULL)
    return NULL;

  if (strcmp (ptr, ".tar") != 0)
    return NULL;

  if ((tarfile = fopen (this->binfile, "rb")) == NULL) {
    uuerror = FILEERR;
    return NULL;
  }

  while (!feof (tarfile)) {
    if (fread (&thisblock, 1, TARBLOCK, tarfile) != TARBLOCK) {
      fclose (tarfile);
      return result;
    }
    if (feof (tarfile)) {
      fclose (tarfile);
      return result;
    }
    if (thisblock.dummy[0] == '\0') {
      fclose (tarfile);
      return result;
    }
    if (thisblock.header.name[0] == '\0') {
      blockno++;
      continue;
    }
    if (thisblock.header.linkflag == '\0' ||
	thisblock.header.linkflag ==  0) {
      blockno++;
      continue;
    }
    if ((new = (uulist *) malloc (sizeof (uulist))) == NULL) {
      fclose (tarfile);
      killlist (result);
      this->branch = NULL;
      uuerror = NOMEM;
      return NULL;
    }
    new->state     = ST_DECODED;
    new->mode      = (short) getokt (thisblock.header.mode);
    new->filename  = _CopyLine (thisblock.header.name, -1);
    new->subfname  = NULL;
    new->tarblock  = blockno++;
    new->binfile   = _CopyLine (this->binfile, -1);
    new->thisfile  = NULL;
    new->misparts  = NULL;
    new->branch    = NULL;
    new->NEXT      = NULL;

    if (result == NULL) 
      result = this->branch = iter = new;
    else {
      iter->NEXT = new;
      iter       = new;
    }

    fsize = getokt (thisblock.header.size);

    while (fsize > 0 && !feof (tarfile)) {
      if (fread (&thisblock, 1, TARBLOCK, tarfile) != TARBLOCK) {
        fclose (tarfile);
        return result;
      }
      fsize -= TARBLOCK;
      blockno++;
    }
  }
  fclose (tarfile);

  return result;
}

