#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include "uudeview.h"
#include "uuerror.h"
#include "uucheck.h"

/*
/* Simply copy a line up to a CR or the end of the string
/**/

char *_CopyLine (char *ptr, int maxlen)
{
  char *result;

  if (maxlen == -1) {
    while (*ptr == ' ') ptr++;
    for (maxlen=0; ptr[maxlen]&&ptr[maxlen]!=0x0a&&ptr[maxlen]!=0x0d; maxlen++);
    while (maxlen>0 && ptr[maxlen-1] == ' ')
      maxlen--;
  }

  if ((result = (char *) malloc (maxlen + 1)) == NULL) {
    uuerror = NOMEM;
    return NULL;
  }

  memcpy (result, ptr, maxlen);
  result[maxlen] = '\0';

  return result;
}

char *cutdir (char *filename)
{
  char *ptr;

  if (filename == NULL)
    return NULL;

  if ((ptr = strrchr (filename, '/')) == NULL)
    ptr = filename;
  else
    ptr++;

  return _CopyLine (ptr, -1);
}

char *dosfname (char *filename)
{
  char dosname[13], *ptr1, *ptr2;
  int count = 0;

  ptr1 = dosname;
  ptr2 = filename;

  while (*ptr2 && *ptr2 != '.' && count++ < 8) {
    if (*ptr2 == ' ') {
      *ptr1++ = '_';
      ptr2++;
    }
    else 
      *ptr1++ = *ptr2++;
  }

  *ptr1++ = '.';
  count = 0;

  if (*ptr2++ == '.') {
    while (*ptr2 && *ptr2 != '.' && count++ < 3) {
      if (*ptr2 == ' ') {
        *ptr1++ = '_';
        ptr2++;
      }
      else
        *ptr1++ = *ptr2++;
    }
  }
  *ptr1 = '\0';

  return _CopyLine (dosname, -1);
}

char *filemode (int mode)
{
  static char result[11];
  static char rwxkey[8][4] = 
    { "---", "--x", "-w-", "-wx", 
      "r--", "r-x", "rw-", "rwx" };

  result[0] = '-';
  result[1] = rwxkey[((mode >> 6) & 0x07)][0];
  result[2] = rwxkey[((mode >> 6) & 0x07)][1];
  result[3] = rwxkey[((mode >> 6) & 0x07)][2];
  result[4] = rwxkey[((mode >> 3) & 0x07)][0];
  result[5] = rwxkey[((mode >> 3) & 0x07)][1];
  result[6] = rwxkey[((mode >> 3) & 0x07)][2];
  result[7] = rwxkey[  mode       & 0x07 ][0];
  result[8] = rwxkey[  mode       & 0x07 ][1];
  result[9] = rwxkey[  mode       & 0x07 ][2];
  result[10]= '\0';

  return result;
}

/*
/* Insert one part of a file into the global list
/**/

int InsertPartToList (uufile *data)
{
  uulist *iter = GlobalFileList, *new;
  uufile *fiter, *last;

  while (iter != NULL) {
    if (strcmp (data->subfname, iter->subfname) == 0) {
      if (iter->filename == NULL && data->filename != NULL) {
        if ((iter->filename = (char *) malloc (strlen (data->filename) + 1)) == NULL) {
	  uuerror = NOMEM;
          return 0;
        }
	strcpy (iter->filename, data->filename);
      }
      if (iter->mode == 0 && data->data->mode != 0)
        iter->mode = data->data->mode;
      if (data->data->uudet)
        iter->uudet = 1;
      if (data->data->begin)
        iter->begin = 1;
      if (data->data->end)
        iter->end = 1;       

      if (data->partno < iter->thisfile->partno) {
	data->NEXT     = iter->thisfile;
	iter->thisfile = data;
	
	return 1;
      }
      fiter = iter->thisfile;
      last  = NULL;

      while (fiter) {
	if (data->partno == fiter->partno) {
          if (fiter->data->subject == NULL)
            return 0;
          else if (strstr (fiter->data->subject, "REPOST") != NULL ||
                   strstr (fiter->data->subject, "Repost") != NULL ||
                   strstr (fiter->data->subject, "repost") != NULL)
            return 0;
          else if (fiter->data->uudet && !data->data->uudet)
            return 0;
          else {
            data->NEXT  = fiter->NEXT;
            fiter->NEXT = NULL;
            killfile (fiter);

            if (last == NULL)
              iter->thisfile = data;
            else
              last->NEXT     = data;

            return 1;
          }
        }

	if (fiter->NEXT == NULL || data->partno < fiter->NEXT->partno) {
	  data->NEXT  = fiter->NEXT;
	  fiter->NEXT = data;
	  return 1;
	}
        last  = fiter;
	fiter = fiter->NEXT;
      }
      
      return -1; /* Shouldn't get here */
    }
    if (iter->NEXT == NULL) 
      break;

    iter = iter->NEXT;
  }
  if ((new = (uulist *) malloc (sizeof (uulist))) == NULL) {
    uuerror = NOMEM;
    return -1;
  }

  if ((new->subfname = (char *) malloc (strlen (data->subfname) + 1)) == NULL) {
    uuerror = NOMEM;
    free (new);
    return -1;
  }
  if (data->filename != NULL) {
    if ((new->filename = (char *) malloc (strlen (data->filename) + 1)) == NULL) {
      uuerror = NOMEM;
      free (new->subfname);
      free (new);
      return -1;
    }
    strcpy (new->filename, data->filename);
  }
  else
    new->filename = NULL;

  new->state    = ST_READ;
  new->binfile  = NULL;
  new->tarblock = -1;
  new->thisfile = data;
  new->mode     = data->data->mode;
  new->uudet    = data->data->uudet;
  new->begin    = data->data->begin;
  new->end      = data->data->end;
  new->misparts = NULL;
  new->NEXT     = NULL;
  new->branch   = NULL;

  strcpy (new->subfname, data->subfname);

  if (iter == NULL)
    GlobalFileList = new;
  else
    iter->NEXT = new;

  return 1;
}

long exists (char *filename)
{
  FILE *datei;
  long pos;

  if (filename == NULL || *filename == '\0')
    return 0;

  if ((datei = fopen (filename, "rb")) == NULL)
    return 0;

  fseek (datei, 0, SEEK_END);
  pos = ftell (datei);
  
  fclose (datei);
  return pos;
}

int iscomp (uulist *data)
{
  char *ptr;

  if ((ptr = strrchr (data->filename, '.')) == NULL)
    return 0;

  if (strcmp (ptr, ".Z"  ) == 0 || strcmp (ptr, ".tz" ) == 0 ||
      strcmp (ptr, ".z"  ) == 0 || strcmp (ptr, ".gz" ) == 0 ||
      strcmp (ptr, ".tgz") == 0 || strcmp (ptr, ".taz") == 0)
    return 1;

  return 0;
}

long getokt (char *ptr)
{
  long result = 0;

  while (*ptr && isdigit(*ptr))
    result = result * 8 + *ptr++ - '0';

  return result;
}

int istarfile (uulist *this)
{
  char *ptr;

  if (this == NULL)
    return 0;

  if ((ptr = strrchr (this->filename, '.')) == NULL)
    return 0;

  if (strcmp (ptr, ".tar") == 0)
    return 1;
  else
    return 0;
}

int findpart (uulist *this, short partno)
{
  uufile *iter = this->thisfile;

  while (iter) {
    if (iter->partno == partno)
      return 1;
    iter = iter->NEXT;
  }
  return 0;
}

uulist *_smparts_r (uulist *addit)
{
  uulist *iter = GlobalFileList;
  uufile *fiter, *dest, *temp;
  int count, flag, copy;

  while (iter) {
    if (iter->state >= ST_OK || iter->uudet == 0) {
      iter = iter->NEXT;
      continue;
    }
    if (iter == addit) {
      iter = iter->NEXT;
      continue;
    }
    if ((iter->begin && addit->begin) || (iter->end && addit->end)) {
      iter = iter->NEXT;
      continue;
    }
    flag  = 0;
    fiter = addit->thisfile;

    while (fiter) {
      if (!(fiter->data->uudet)) {
	fiter = fiter->NEXT;
	continue;
      }

      if (findpart (iter, fiter->partno)) {
	flag = 0;
	break;
      }
      else 
	flag = 1;

      fiter = fiter->NEXT;
    }
    if (flag == 0) {
      iter = iter->NEXT;
      continue;
    }
    dest  = iter->thisfile;
    fiter = addit->thisfile;

    if (iter->filename == NULL && addit->filename != NULL) {
      if ((iter->filename = (char *) malloc (strlen (addit->filename) + 1)) == NULL) {
        uuerror = NOMEM;
        return 0;
      }
      strcpy (iter->filename, addit->filename);
    }

    if (addit->begin) iter->begin = 1;
    if (addit->end)   iter->end   = 1;

    if (addit->mode != 0 && iter->mode == 0)
      iter->mode = addit->mode;

    while (fiter) {
      flag = 0;

      if (fiter->partno == iter->thisfile->partno ||
	  (dest->NEXT != NULL && fiter->partno == dest->NEXT->partno)) {
	temp           = fiter->NEXT;
	fiter->NEXT    = NULL;

	killfile (fiter);

	addit->thisfile= temp;
	fiter          = temp;
	continue;
      }
      if (fiter->partno < iter->thisfile->partno) {
	temp           = fiter->NEXT;
	fiter->NEXT    = iter->thisfile;
	iter->thisfile = fiter;
	dest           = fiter;
	addit->thisfile= temp;
	fiter          = temp;
      }
      else if (dest->NEXT == NULL || fiter->partno < dest->NEXT->partno) {
	temp           = fiter->NEXT;
	fiter->NEXT    = dest->NEXT;
	dest->NEXT     = fiter;
	addit->thisfile= temp;
	fiter          = temp;
      }
      else {
	dest = dest->NEXT;
      }
    }
    break;
  }
  return iter;
}

uulist *smergeparts (void)
{
  uulist *iter = GlobalFileList, *last=NULL, *res, *temp;
  int flag = 0, pflag=0;
  uufile *fiter;

  while (iter) {
    if (iter->state >= ST_OK || iter->uudet == 0) {
      last = iter;
      iter = iter->NEXT;
      continue;
    }
    if ((res = _smparts_r (iter)) != NULL) {
      if (pflag++ == 0)
	printf ("\n");

      printf ("Parts of '%s' merged with parts of '%s'. Check ",
	      (iter->subfname) ? iter->subfname : "",
	      (res->subfname)  ? res->subfname  : "");
 
      temp       = iter->NEXT;
      iter->NEXT = NULL;
      killlist (iter);

      /* Recheck _res_, maybe it's OK now */

      if (checkagain (res)->state == ST_OK)
	printf ("OK");

      printf ("\n");

      if (last == NULL) {
	GlobalFileList = temp;
	iter           = temp;
      }
      else {
	last->NEXT     = temp;
	iter           = temp;
      }

      continue;
    }
    last = iter;
    iter = iter->NEXT;
  }
  if (pflag)
    printf ("\n");

  return GlobalFileList;
}

/*****************************************************************************
 + Frank Pilhofer                          frank@informatik.uni-frankfurt.de +
 +---------------------------------------------------------------------------+
 | Department of Computer Sciences * University of Frankfurt / Main, Germany |
 *****************************************************************************/
